<?php

use Carbon\Carbon;
use Elementor\Plugin;
use PRAMADILLO\Woocommerce_Pay_Per_Post_Restrict_Content;


/**
 * Class Woocommerce_Pay_Per_Post_General
 */
class Woocommerce_Pay_Per_Post_Helper extends Woocommerce_Pay_Per_Post {

	/**
	 * @var array
	 */
	public static $protection_types = [ 'standard', 'delay', 'page-view', 'expire' ];

	/**
	 * @param $post_id
	 *
	 * @return bool|string
	 */
	public static function is_protected( $post_id = null ) {

		if ( is_null( $post_id ) ) {
			$post_id = get_the_ID();
		}

		$selected = (array) get_post_meta( $post_id, WC_PPP_SLUG . '_product_ids' );
		$selected = array_filter( $selected );

		//Elementor Integration
		if ( wcppp_freemius()->is__premium_only() ) {
			if ( wcppp_freemius()->can_use_premium_code() ) {
				if ( Woocommerce_Pay_Per_Post_Helper::can_use_elementor() ) {
					$elementor_data = json_decode( (string) get_post_meta( $post_id, '_elementor_data', true ) );
					if ( ! is_null( $elementor_data ) ) {
						$product_ids = Woocommerce_Pay_Per_Post_Helper::recursive_array_search( 'wc_pay_per_post_select_products', $elementor_data );
						if ( $product_ids && count( $product_ids ) > 0 ) {
							$selected = array_merge( $selected, $product_ids );
						}
					}
				}
			}
		}

		$delay_restriction_enable     = (bool) get_post_meta( $post_id, WC_PPP_SLUG . '_delay_restriction_enable', true );
		$page_view_restriction_enable = (bool) get_post_meta( $post_id, WC_PPP_SLUG . '_page_view_restriction_enable', true );
		$expire_restriction_enable    = (bool) get_post_meta( $post_id, WC_PPP_SLUG . '_expire_restriction_enable', true );

		if ( ! empty( $selected ) ) {
			$protection = 'standard';

			if ( $delay_restriction_enable ) {
				$protection = 'delay';
			}

			if ( $page_view_restriction_enable ) {
				$protection = 'page-view';
			}

			if ( $expire_restriction_enable ) {
				$protection = 'expire';
			}

			return $protection;

		} else {
			return false;
		}

	}

	/**
	 * @param null $post_id
	 * @param bool $track_page_view *
	 *
	 * @return bool
	 * The can_user_view_content function returns on whether or not the user should see the paywall.
	 * For this that is why we are returning the inverse of the result.
	 */
	public static function has_access( $post_id = null, bool $track_page_view = true ): bool {
		$restrict = new Woocommerce_Pay_Per_Post_Restrict_Content( $post_id, $track_page_view );

		return ! $restrict->can_user_view_content();
	}

	/**
	 * Checks if the user has purchased a product associated with the post
	 *
	 * @param null $post_id
	 * @param bool $track_page_view
	 *
	 * @return bool
	 */
	public static function has_purchased( $post_id = null, bool $track_page_view = true ): bool {
		$restrict = new Woocommerce_Pay_Per_Post_Restrict_Content( $post_id, $track_page_view );

		return $restrict->check_if_purchased();
	}

	/**
	 * @return string
	 */
	public static function get_no_access_content(): string {
		$restrict = new Woocommerce_Pay_Per_Post_Restrict_Content();

		return $restrict->show_paywall( get_the_content() );
	}

	/**
	 * @param $type
	 *
	 * @return bool|string
	 */
	public static function protection_display_icon( $type ) {

		if ( in_array( $type, self::$protection_types, true ) ) {

			switch ( $type ) {
				case 'standard':
					return '<span class="dashicons dashicons-post-status" title="Standard Purchase Protection" style="color:green"></span>';
				case 'delay':
					return '<span class="dashicons dashicons-clock" title="Delay Protection" style="color:green"></span>';
				case 'page-view':
					return '<span class="dashicons dashicons-visibility" title="Page View Protection" style="color:green"></span>';
				case 'expire':
					return '<span class="dashicons dashicons-backup" title="Expiry Protection" style="color:green"></span>';
			}
		}

		return false;

	}

	/**
	 * @return Carbon
	 */
	public static function current_time(): Carbon {
		return Carbon::createFromTimestamp( current_time( 'timestamp' ) );
	}

	/**
	 * @return string
	 */
	public static function date_time_format(): string {
		return get_option( 'date_format', true ) . ' ' . get_option( 'time_format', true );
	}

	public static function logger( $message, $context = [] ) {
		$logger = new Woocommerce_Pay_Per_Post_Logger();
		$logger->log( $message, $context );
	}

	public static function logger_uri(): string {
		$logger = new Woocommerce_Pay_Per_Post_Logger();

		return $logger->get_log_uri();
	}

	public static function logger_url(): string {
		$logger = new Woocommerce_Pay_Per_Post_Logger();

		return $logger->get_log_url();
	}

	public static function get_protected_posts( $args = null, $transient = 'posts' ): array {

		$transient = WC_PPP_SLUG . '_' . $transient;

		if ( is_null( $args ) ) {
			$custom_post_types = get_option( WC_PPP_SLUG . '_custom_post_types', [] );
			$custom_post_types = empty( $custom_post_types ) ? [] : $custom_post_types;

			if ( ! is_array( $custom_post_types ) ) {
				$custom_post_types = explode( ',', $custom_post_types );
			}

			$args = [
				'orderby'  => 'post_date',
				'order'    => 'DESC',
				'nopaging' => true,

				'meta_query'  => [
					'relation' => 'OR',
					[
						'key'     => WC_PPP_SLUG . '_product_ids',
						'value'   => '',
						'compare' => '!=',
					],
					[
						'key'     => '_elementor_data',
						'value'   => '"wc_pay_per_post_enable":"yes"', //TODO There has to be a better way to do this.
						'compare' => 'LIKE',
					],
				],
				'post_status' => 'publish',
				'post_type'   => $custom_post_types,
			];
		}

		$ppp_posts = get_transient( $transient );
		if ( false === $ppp_posts ) {
			$ppp_posts = get_posts( $args );
			set_transient( $transient, $ppp_posts, apply_filters( 'wc_pay_per_post_posts_transient_time', DAY_IN_SECONDS ) );
		}

		return (array) $ppp_posts;
	}

	public static function get_posts_associated_with_product_id( $product_id ): array {
		$custom_post_types = get_option( WC_PPP_SLUG . '_custom_post_types', [] );
		$custom_post_types = empty( $custom_post_types ) ? [] : $custom_post_types;

		if ( ! is_array( $custom_post_types ) ) {
			$custom_post_types = explode( ',', $custom_post_types );
		}

		$args = [
			'orderby'    => 'post_date',
			'order'      => 'DESC',
			'nopaging'   => true,
			'meta_query' => [
				'relation' => 'OR',
				[
					'key'     => WC_PPP_SLUG . '_product_ids',
					'value'   => sprintf( '^%1$s$|s:%2$u:"%1$s";', $product_id, strlen( $product_id ) ),
					'compare' => 'REGEXP',
				],
				[
					'key'     => '_elementor_data',
					'value'   => '"' . $product_id . '"',
					'compare' => 'LIKE',
				],
			],

			'post_status' => 'publish',
			'post_type'   => $custom_post_types,
		];

		$posts = get_posts( apply_filters( 'wc_pay_per_post_woocommerce_email_args', $args ) );

		$protected_content = [];

		if ( $posts ) {
			$interface = 'default';

			if ( wcppp_freemius()->is__premium_only() ) {
				if ( wcppp_freemius()->can_use_premium_code() ) {

					if ( function_exists( 'pll_current_language' ) && get_option( WC_PPP_SLUG . '_enable_polylang', false ) ) {
						$interface = 'polylang';
					}

				}
			}

			switch ( $interface ) {
				case "default":
				default:
					$protected_content = self::get_posts_associated_with_product_id_interface( $posts );
					break;

				case "polylang":
					$protected_content = self::get_posts_associated_with_product_id_interface_polylang__premium_only( $posts );
					break;
			}

		}

		return $protected_content;

	}

	public static function can_use_woocommerce_memberships(): bool {
		if ( wcppp_freemius()->is__premium_only() ) {
			if ( wcppp_freemius()->can_use_premium_code() ) {
				if ( is_plugin_active( 'woocommerce-memberships/woocommerce-memberships.php' ) ) {
					return true;
				}
			}
		}

		return false;
	}

	public static function can_use_woocommerce_subscriptions(): bool {
		if ( wcppp_freemius()->is__premium_only() ) {
			if ( wcppp_freemius()->can_use_premium_code() ) {
				if ( is_plugin_active( 'woocommerce-subscriptions/woocommerce-subscriptions.php' ) ) {
					return true;
				}
			}
		}

		return false;
	}

	public static function can_use_paid_membership_pro(): bool {
		if ( wcppp_freemius()->is__premium_only() ) {
			if ( wcppp_freemius()->can_use_premium_code() ) {
				if ( is_plugin_active( 'paid-memberships-pro/paid-memberships-pro.php' ) ) {
					return true;
				}
			}
		}

		return false;
	}

	public static function can_use_elementor(): bool {

		if ( wcppp_freemius()->is__premium_only() ) {
			if ( wcppp_freemius()->can_use_premium_code() ) {
				if ( is_plugin_active( 'elementor/elementor.php' ) ) {
					return true;
				}
			}
		}

		return false;
	}

	public static function can_use_polylang__premium_only(): bool {
		if ( function_exists( 'pll_current_language' ) && get_option( WC_PPP_SLUG . '_enable_polylang', false ) ) {
			return true;

		}

		return false;
	}

	public static function md_array_diff( $arraya, $arrayb ) {

		foreach ( $arraya as $keya => $valuea ) {
			if ( in_array( $valuea, $arrayb ) ) {
				unset( $arraya[ $keya ] );
			}
		}

		return $arraya;
	}

	public static function move_to_top( &$array, $key ) {
		$temp = [ $key => $array[ $key ] ];
		unset( $array[ $key ] );
		$array = $temp + $array;
	}

	public static function get_the_excerpt( $post_id ): string {
		return get_the_excerpt( $post_id );
	}

	protected static function get_posts_associated_with_product_id_interface_polylang__premium_only( $posts ): array {
		$translations      = [];
		$protected_content = [];
		foreach ( $posts as $post ) {
			$get_translations = get_the_terms( $post->ID, 'post_translations' );

			if ( ! in_array( $post->ID, $translations ) ) {
				$protected_content[] = [
					'post_id'      => $post->ID,
					'post_title'   => $post->post_title,
					'post_url'     => get_permalink( $post->ID ),
					'translations' => $translations,
				];
			}

			if ( $get_translations ) {
				$translations = unserialize( $get_translations[0]->description );
			}

		}

		return $protected_content;
	}

	protected static function get_posts_associated_with_product_id_interface( $posts ): array {
		$protected_content = [];

		foreach ( $posts as $post ) {
			$protected_content[] = [
				'post_id'    => $post->ID,
				'post_title' => $post->post_title,
				'post_url'   => get_permalink( $post->ID ),
			];
		}

		return $protected_content;
	}

	public static function locate_template( $template_name, $template_path = '', $default_path = '' ) {
		if ( ! $template_path ) {
			$template_path = WC_PPP_TEMPLATE_PATH;
		}

		if ( ! $default_path ) {
			$default_path = WC_PPP_PATH;
		}

		// Look within passed path within the theme - this is priority.
		$template = locate_template(
			[
				trailingslashit( $template_path ) . $template_name,
				$template_name,
			]
		);

		// Get default template/.
		if ( ! $template ) {
			$template = $default_path . $template_name;
		}

		// Return what we found.
		return apply_filters( 'wc_pay_for_post_locate_template', $template, $template_name, $template_path );
	}

	public static function get_all_products(): array {
		$args = [
			'post_type'   => [ 'product' ],
			'orderby'     => 'title',
			'post_status' => 'publish',
			'order'       => 'ASC',
			'nopaging'    => true,
		];

		if ( wcppp_freemius()->is__premium_only() ) {
			if ( wcppp_freemius()->can_use_premium_code() ) {
				$args['post_type'] = [ 'product', 'product_variation' ];
			}
		}

		$products = get_posts( apply_filters( 'wc_pay_per_post_all_product_args', $args ) ); //TODO: Document this filter

		$return = [];

		foreach ( $products as $product ) {
			$return[] = [
				'ID'         => $product->ID,
				'post_title' => $product->post_title,
			];
		}

		return $return;
	}

	public static function get_virtual_products(): array { //TODO Check/test this. Document in 3.0.0
		$args = [
			'post_type'   => [ 'product' ],
			'post_status' => 'publish',
			'orderby'     => 'title',
			'order'       => 'ASC',
			'nopaging'    => true,
			'meta_query'  => [
				'relation' => 'OR',
				[
					'key'     => '_downloadable',
					'value'   => 'yes',
					'compare' => '=',
				],
				[
					'key'     => '_virtual',
					'value'   => 'yes',
					'compare' => '=',
				],
				[
					'key'     => '_downloadable',
					'value'   => '1',
					'compare' => '=',
				],
				[
					'key'     => '_virtual',
					'value'   => '1',
					'compare' => '=',
				],
			],

		];

		if ( wcppp_freemius()->is__premium_only() ) {
			if ( wcppp_freemius()->can_use_premium_code() ) {
				$args['post_type'] = [ 'product', 'product_variation' ];
			}
		}

		$products = get_posts( apply_filters( 'wc_pay_per_post_virtual_product_args', $args ) );

		$return = [];

		foreach ( $products as $product ) {
			$return[] = [
				'ID'         => $product->ID,
				'post_title' => $product->post_title,
			];
		}

		return $return;
	}

	public static function get_products() {
		$only_show_virtual_products = (bool) get_option( WC_PPP_SLUG . '_only_show_virtual_products', false );

		if ( $only_show_virtual_products ) {
			return apply_filters( 'wc_pay_per_post_get_virtual_products', Woocommerce_Pay_Per_Post_Helper::get_virtual_products() );
		} else {
			return apply_filters( 'wc_pay_per_post_get_all_products', Woocommerce_Pay_Per_Post_Helper::get_all_products() );
		}
	}

	public static function replace_tokens( $paywall_content, $product_ids, $unfiltered_content = null ) {
		$excerpt   = apply_filters( 'wc_pay_per_post_modify_excerpt', wp_trim_words( $unfiltered_content ) );
		$parent_id = null;

		if ( isset( $product_ids[0] ) ) {
			$parent_id = wp_get_post_parent_id( $product_ids[0] );
		}

		if ( is_archive() || is_home() || is_front_page() ) {
			//Get Product IDs
			$product_ids = get_post_meta( get_the_ID(), WC_PPP_SLUG . '_product_ids', true );
		}

		$return_content = str_replace( '{{product_id}}', implode( ',', (array) $product_ids ), $paywall_content );
		$return_content = str_replace( '{{parent_id}}', $parent_id, $return_content );

		if ( wcppp_freemius()->is__premium_only() ) {
			if ( wcppp_freemius()->can_use_premium_code() ) {
				$return_content = str_replace( '{{excerpt}}', $excerpt, $return_content );
			}
		}

		return $return_content;
	}

	/**
	 * @param $method
	 *
	 * @return string
	 */
	public static function carbon_add_method( $method ): string {
		$methods = [
			'minute' => 'addMinutes',
			'hour'   => 'addHours',
			'day'    => 'addDays',
			'week'   => 'addWeeks',
			'month'  => 'addMonths',
			'year'   => 'addYears',
		];

		return $methods[ $method ];
	}

	/**
	 * @param $method
	 *
	 * @return string
	 */
	public static function carbon_diff_method( $method ): string {
		$methods = [
			'minute' => 'diffInMinutes',
			'hour'   => 'diffInHours',
			'day'    => 'diffInDays',
			'week'   => 'diffInWeeks',
			'month'  => 'diffInMonths',
			'year'   => 'diffInYears',
		];

		return $methods[ $method ];
	}

	public static function recursive_array_search( $needle, array $haystack ) {
		$matches = [];

		$iterator  = new RecursiveArrayIterator( $haystack );
		$recursive = new RecursiveIteratorIterator(
			$iterator,
			RecursiveIteratorIterator::SELF_FIRST
		);

		foreach ( $recursive as $key => $value ) {
			if ( $key === $needle ) {
				$matches[] = $value;
			}
		}
		$return = [];
		foreach ( $matches as $match => $mvalue ) {
			$return = array_merge( $return, $mvalue );
		}

		return ( $return ) ?? false;
	}

	public static function flatten( $array ): array {
		$return = [];
		foreach ( $array as $key => $value ) {
			$return[] = $value[0];
		}

		return $return;
	}

	public static function array_values_recursive( $array ): array {
		$flat = array();

		foreach ( $array as $value ) {
			if ( is_array( $value ) ) {
				$flat = array_merge( $flat, self::array_values_recursive( $value ) );
			} else {
				$flat[] = $value;
			}
		}

		return $flat;
	}

	public static function get_elementor_product_ids( $post_id ): array {
		$product_ids    = [];
		$elementor_data = get_post_meta( $post_id, '_elementor_data', true );
//            var_dump($elementor_data);
		if ( ! is_null( $elementor_data ) && false !== strpos( (string) $elementor_data, '"wc_pay_per_post_enable":"yes"' ) ) {

			$document      = Plugin::$instance->documents->get( $post_id );
			$product_ids[] = Plugin::$instance->db->iterate_data( $document->get_elements_data(), function ( $element_data ) {
				$element = Plugin::$instance->elements_manager->create_element_instance( $element_data );
				// If the widget/element isn't exist, like a plugin that creates a widget but deactivated
				if ( ! $element ) {
					return null;
				}

				$data = $element->get_data();
				if ( false !== strpos( json_encode( $data ), '"wc_pay_per_post_enable":"yes"' ) ) {
					return Woocommerce_Pay_Per_Post_Helper::recursive_array_search( 'wc_pay_per_post_select_products', $data );
				}

			} );
		}

		return self::array_values_recursive( $product_ids );
	}


	public static function get_post_types(): array {

		$user_included_post_types = get_option( WC_PPP_SLUG . '_custom_post_types', [] );

		if ( '' === $user_included_post_types || empty( $user_included_post_types ) ) {
			$user_included_post_types = [];
		}

		return (array) $user_included_post_types;

	}

}
